if not modules then modules = { } end modules ['mcallbacks'] = {
    version   = 0.99,
    comment   = "simple module to have several functions in a callback",
    author    = "Hans Hagen & Elie Roux",
    copyright = "Hans Hagen & Elie Roux",
    license   = "public domain",
}

if mcallbacks and mcallbacks.version then
 -- we simply don't load
else

mcallbacks          = mcallbacks or { }
mcallbacks.version  = 0.99
mcallbacks.showlog  = mcallbacks.showlog or false

mcallbacks.callbacklist = mcallbacks.callbacklist or { }

local list = 1
local data = 2

local format = string.format

mcallbacks.callbacktypes = mcallbacks.callbacktypes or {
hpack_filter = list,
pre_linebreak_filter = list,
post_linebreak_filter = data,
}

-- as we will overwrite callback.register later, we keep it in an "internal" function
mcallbacks.internalregister = mcallbacks.internalregister or callback.register

local callbacktypes = mcallbacks.callbacktypes

mcallbacks.report = mcallbacks.report or function(...)
  texio.write_nl(format("(mcallbacks: %s)",string.format(...)))
end

mcallbacks.log = mcallbacks.log or function(...)
  if mcallbacks.showlog then
    mcallbacks.report(...)
  end
end

mcallbacks.error = mcallbacks.error or function(...)
  mcallbacks.report(...)
end

mcallbacks.add = mcallbacks.add or function (name,description,func,priority)
    if type(func) ~= "function" then
        mcallbacks.error("unable to add function, no proper function passed")
        return
    end
    if not name or name == "" then
        mcallbacks.error("unable to add function, no proper callback name passed")
        return
    elseif not callbacktypes[name] then
        mcallbacks.error(format("unable to add function, %s is not a valid callback", name))
        return
    end
    
    if not description or description == "" then
        mcallbacks.error(format("unable to add function to '%s', no proper description passed",name))
        return
    end
    local l = mcallbacks.callbacklist[name]
    if not l then
        l = { }
        mcallbacks.callbacklist[name] = l
        if callbacktypes[name] == list then
            mcallbacks.internalregister(name, mcallbacks.listhandler(name))
        elseif callbacktypes[name] == data then
            mcallbacks.internalregister(name, mcallbacks.listhandler(name))
        else
            mcallbacks.error("unknown callback type")
        end
        mcallbacks.log(format("creating callback list for '%s'",name))
    end
    local f = {
        func = func,
        description = description,
    }
    priority = tonumber(priority)
    if not priority or priority > #l then
        priority = #l+1
    elseif priority < 1 then
        priority = 1
    end
    table.insert(l,priority,f)
    mcallbacks.log(format("inserting function '%s' at position %s in callback list for '%s'", description,priority,name))
end

mcallbacks.remove = mcallbacks.remove or function (name,description)
    if not name or name == "" then
        mcallbacks.error("unable to remove function, no proper callback name passed")
        return
    elseif not callbacktypes[name] then
        mcallbacks.error(format("unable to remove function, %s is not a valid callback", name))
        return
    end
    if not description or description == "" then
        mcallbacks.error(format("unable to remove function from '%s', no proper description passed",name))
        return
    end
    local l = mcallbacks.callbacklist[name]
    if not l then
        mcallbacks.error(format("no callback list for '%s'",name))
        return
    end
    for k,v in ipairs(l) do
        if v.description == description then
            table.remove(l,k)
            mcallbacks.log(format("removing function '%s' from '%s'",description,name))
            return
        end
    end
    mcallbacks.error(format("unable to remove function '%s' from '%s'",description,name))
end

mcallbacks.reset = mcallbacks.reset or function (name)
    local l = mcallbacks.callbacklist[name]
    if not l then
        mcallbacks.error(format("unknown callback list '%s'",name))
    else
        mcallbacks.log(format("resetting callback list '%s'",name))
        mcallbacks.callbacklist[name] = { }
    end
end

-- a few examples of handlers

mcallbacks.listhandler = mcallbacks.listhandler or function (name)
    return function(head,...)
        local l = mcallbacks.callbacklist[name]
        if l then
            local done = false
            for _, f in ipairs(l) do
                head, ok = f.func(head,...)
                if ok then done = true end
            end
            return head, done
        else
            return head, false
        end
    end
end

mcallbacks.datahandler = mcallbacks.datahandler or function (data)
    return function(data,...)
        local l = mcallbacks.callbacklist[name]
        if l then
            for _, f in ipairs(l) do
                data = f.func(data,...)
            end
        end
        return data
    end
end

callback.add = mcallbacks.add
callback.remove = mcallbacks.remove
callback.reset = mcallbacks.reset

-- we overwrite callback.register, so that noone uses it
callback.register = function (...)
mcallbacks.error("function callback.register is considered too dangerous to use and has been deleted by mcallbacks, please use callback.add instead")
end

end
